from enum import Enum

from proto_formatter.comment import CommentParser
from proto_formatter.object_parser import ObjectParser
from proto_formatter.proto import Service
from proto_formatter.proto import ServiceElement
from proto_formatter.util import remove_prefix, remove_suffix


class ServiceParser(ObjectParser):
    def __init__(self):
        obj = Service(name='name', elements=[], comments=[])
        super(ServiceParser, self).__init__(obj=obj)

    @classmethod
    def parse_obj_field(cls, line, top_comments=[]):
        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = line.strip().replace('(', '')
        line = line.replace(')', '')

        semicolon_index = line.index(cls.SEMICOLON)
        str_before_semicolon = line[:semicolon_index]
        parts = str_before_semicolon.split(' ')
        parts = list(filter(None, parts))
        comments = CommentParser.create_comment(line, top_comments)
        return ServiceElement(label=parts[0], name=parts[1], request=parts[2], response=parts[4], comments=comments)



from proto_formatter.proto import Comment
from proto_formatter.comment import CommentParser
from proto_formatter.constant import Constant
from proto_formatter.proto import Position
from proto_formatter.util import remove_prefix, remove_suffix


class Detector(Constant):

    def get_object_type(self, lines):
        comment_parser = CommentParser()
        comments = comment_parser.pick_up_comment(lines)
        comments = comment_parser.parse(comments)

        line = lines[0]
        if self._is_syntax_line(line):
            return 'syntax', comments
        if self._is_package_line(line):
            return 'package', comments
        if self._is_option_line(line):
            return 'option', comments
        if self._is_import_line(line):
            return 'import', comments
        if self._is_message_object(line):
            return 'message', comments
        if self._is_element_line(line):
            return 'element_field', comments
        if self._is_enum_object(line):
            return 'enum', comments
        if self._is_service_object(line):
            return 'service', comments

        return None, None

    def _is_syntax_line(self, line):
        return line.replace(' ', '').startswith('syntax=')

    def _is_package_line(self, line):
        return line.strip().startswith('package ')

    def _is_option_line(self, line):
        return line.strip().startswith('option ')

    def _is_import_line(self, line):
        return line.strip().startswith('import ')

    def _is_object_start(self, line):
        if line.count(self.BRACE_LEFT) == 0:
            return False

        if line.count(self.SINGLE_COMMENT_SYMBOL) > 0:
            if line.index(self.BRACE_LEFT) > line.index(self.SINGLE_COMMENT_SYMBOL):
                return False

        if line.count(self.MULTIPLE_COMENT_START_SYMBOL) > 0:
            if line.index(self.BRACE_LEFT) > line.index(self.MULTIPLE_COMENT_START_SYMBOL):
                return False

        return True

    def _is_object_end(self, line):
        if line.count(self.BRACE_RIGHT) == 0:
            return False

        if line.count(self.SINGLE_COMMENT_SYMBOL) > 0:
            if line.index(self.BRACE_RIGHT) > line.index(self.SINGLE_COMMENT_SYMBOL):
                return False

        if line.count(self.MULTIPLE_COMENT_START_SYMBOL) > 0:
            if line.index(self.BRACE_RIGHT) > line.index(self.MULTIPLE_COMENT_START_SYMBOL):
                return False

        return True

    def _is_message_object(self, line):
        return line.strip().startswith('message ') and line.strip().count(self.BRACE_LEFT)

    def _is_element_line(self, line):
        if line.count(self.SEMICOLON) == 0:
            return False

        if line.count(self.SINGLE_COMMENT_SYMBOL) > 0:
            if line.index(self.SEMICOLON) > line.index(self.SINGLE_COMMENT_SYMBOL):
                return False

        if line.count(self.MULTIPLE_COMENT_START_SYMBOL) > 0:
            if ine.index(self.SEMICOLON) > ine.index(self.MULTIPLE_COMENT_START_SYMBOL):
                return False

        if self._is_service_element_line(line):
            return True

        if self._is_map_element_line(line):
            return True

        return line.strip().count(self.SEMICOLON) > 0 and line.strip().count(self.EQUAL_SIGN) > 0

    def _is_service_element_line(self, line):
        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = line.strip()
        return line.startswith('rpc ')

    def _is_map_element_line(self, line):
        # map<string, Project> projects = 3;
        line = line.strip().replace(' ', '')
        return line.startswith('map<')

    def _is_enum_object(self, line):
        return line.strip().startswith('enum ') and line.strip().count(self.BRACE_LEFT)

    def _is_service_object(self, line):
        return line.strip().startswith('service ') and line.strip().count(self.BRACE_LEFT)



from proto_formatter.formatter import Formatter
from proto_formatter.parser import ProtoParser


def format_file(fp, indents=2, all_top_comments=False, equal_sign=False, new_fp=None):
    """
    Format protobuf file, override the original file if no new file path(new_fp) specified
    or write the formatted result to a new file.
    :param fp: unformatted protobuf file path.
    :param indents: indents number.
    :param all_top_comments: convert the single line comment at the left of code as a top comment or not. Default: not
    :param equal_sign: if align the code by equal sign or not.
    Example of align:
        ENUM_TYPE_UNSPECIFIED = 0;  // ENUM_TYPE_UNSPECIFIED
        ENUM_TYPE_CARRY_A     = 1;  // ENUM_TYPE_CARRY_A
        ENUM_TYPE_B           = 2;  // ENUM_TYPE_B
    :param new_fp: the file path of new formatted protobuf file. Rewrite the original file if it is not specified.
    :return: file content size.
    """
    parser = ProtoParser()
    protobuf_obj = parser.load(fp=fp)
    formatter = Formatter(indents=indents, all_top_comments=all_top_comments, equal_sign=equal_sign)
    content = formatter.to_string(protobuf_obj)

    if new_fp:
        fp = new_fp

    with open(fp, 'w') as f:
        return f.write(content)

def format_str(proto_str, indents=2, all_top_comments=False, equal_sign=False):
    """
    Format a protobuf string, return the formatted string.
    :param proto_str: protobuf string need to be formatted.
    :param indents: indents number.
    :param all_top_comments: convert the single line comment at the left of code as a top comment or not. Default: not
    :param equal_sign: if align the code by equal sign or not.
    Example of align:
        ENUM_TYPE_UNSPECIFIED = 0;  // ENUM_TYPE_UNSPECIFIED
        ENUM_TYPE_CARRY_A     = 1;  // ENUM_TYPE_CARRY_A
        ENUM_TYPE_B           = 2;  // ENUM_TYPE_B
    :return: formatted string.
    """
    parser = ProtoParser()
    protobuf_obj = parser.loads(proto_str=proto_str.strip())
    formatter = Formatter(indents=indents, all_top_comments=all_top_comments, equal_sign=equal_sign)

    return formatter.to_string(protobuf_obj)




from enum import Enum
from proto_formatter.detector import Detector
from proto_formatter.comment import CommentParser
from proto_formatter.constant import Constant
from proto_formatter.proto import ProtoBufStructure
from proto_formatter.util import remove_prefix, remove_suffix


class ObjectParser(Constant):

    def __init__(self, obj):
        self.objects = []
        self.cursor = 0

        self.left_brace_stack = []
        self.right_brace_stack = []

        # why elements is shared in multiple instanses when Message(name='name')
        self.obj = obj
        self.debug = 0

    @classmethod
    def _get_obj_name(cls, line):
        parts = line.strip().split(' ')
        parts = list(filter(None, parts))
        name = parts[1]
        return name

    @classmethod
    def parse_obj_field(cls, line, top_comments):
        raise NotImplementedError("Element parse not implemented.")

    def parse_and_add(self, proto_obj: ProtoBufStructure, lines, top_comment_list):
        self.parse(lines, top_comment_list, True)
        proto_obj.objects.append(self.obj)

    def parse(self, lines, top_comments, is_object_start):
        if is_object_start:
            current_line = lines.pop(0)
            name = self._get_obj_name(current_line)
            self.obj.name = name
            self.obj.comments = CommentParser.create_comment(current_line, top_comments)
            return self.parse(lines, [], False)

        parser, comments = Detector().get_object_type(lines)
        current_line = lines.pop(0)
        if Detector()._is_object_end(current_line):
            return

        element = self.parse_obj_field(current_line, comments)
        self.obj.elements.append(element)

        return self.parse(lines, [], False)



from proto_formatter.comment import CommentParser
from proto_formatter.proto import Message
from proto_formatter.proto import MessageElement
from proto_formatter.object_parser import ObjectParser
from proto_formatter.proto import ProtoBufStructure
from proto_formatter.proto import Syntax


class MessageParser(ObjectParser):

    def __init__(self):
        obj = Message(name='name', elements=[], comments=[])
        super(MessageParser, self).__init__(obj=obj)

    def parse_obj_field(self, line, top_comments=[]):
        if 'map<' in line:
            return self.make_map_field_element(line, top_comments)

        line = line.strip()
        equal_sign_index = line.index(self.EQUAL_SIGN)
        semicolon_index = line.index(self.SEMICOLON)
        str_before_equqal_sign = line[:equal_sign_index]
        parts = str_before_equqal_sign.split(' ')
        parts = list(filter(None, parts))
        value = line[equal_sign_index + 1:semicolon_index].strip().replace('"', "").replace("'", "")
        comments = CommentParser.create_comment(line, top_comments)
        if len(parts) == 2:
            return MessageElement(label=None, type=parts[0], name=parts[1], number=value, comments=comments)
        if len(parts) == 3:
            return MessageElement(label=parts[0], type=parts[1], name=parts[2], number=value, comments=comments)

        return None

    def make_map_field_element(self, line, top_comments=[]):
        right_bracket_index = line.index(self.ANGLE_BRACKET_RIGHT)
        equal_sign_index = line.index(self.EQUAL_SIGN)
        semicolon_index = line.index(self.SEMICOLON)
        type = line[:right_bracket_index + 1]
        type = type.strip().replace(' ', '')
        type_parts = type.split(',')
        type = ', '.join(type_parts)
        name = line[right_bracket_index + 1:equal_sign_index]
        name = name.strip()
        number = line[equal_sign_index + 1:semicolon_index]
        number = number.strip()
        comments = CommentParser.create_comment(line, top_comments)
        return MessageElement(label=None, type=type, name=name, number=number, comments=comments)



from proto_formatter.proto import EnumElement
from proto_formatter.proto import Import
from proto_formatter.proto import Message
from proto_formatter.proto import MessageElement
from proto_formatter.proto import Option
from proto_formatter.proto import Package
from proto_formatter.proto import Position
from proto_formatter.proto import ProtoBufStructure
from proto_formatter.proto import ProtoEnum
from proto_formatter.proto import Service
from proto_formatter.proto import ServiceElement
from proto_formatter.proto import Syntax


class Formatter():

    def __init__(self, indents=2, equal_sign=None, all_top_comments=False):
        self.indents = indents
        self.equal_sign = equal_sign
        self.all_top_comments = all_top_comments

    def to_string(self, obj: ProtoBufStructure):
        syntax_string = self.syntax_string(obj.syntax)
        package_string = self.package_string(obj.package)
        option_string = self.options_string(obj.options)
        imports_string = self.imports_string(obj.imports)
        messages_string = self.objects_string(obj.objects)
        contents = [syntax_string, package_string, option_string, imports_string, messages_string]
        contents = list(filter(None, contents))
        content = '\n\n'.join(contents)
        content = content + '\n'

        return content

    def syntax_string(self, obj: Syntax):
        if not obj:
            return ''

        line = f'syntax = "{obj.value}";'
        return self.make_string(line, obj.comments)

    def package_string(self, obj: Package):
        if not obj:
            return ''

        line = f'package {obj.value};'
        return self.make_string(line, obj.comments)

    def options_string(self, obj_list):
        if not obj_list:
            return ''

        max_length = self.max_length_of_option(obj_list)

        string_list = []
        for obj in obj_list:
            string = self.option_string(obj, max_length)
            string_list.append(string)
        return '\n'.join(string_list)

    def max_length_of_option(self, obj_list):
        max = 0
        for obj in obj_list:
            if obj.value == 'true' or obj.value == 'false':
                line = f'option {obj.name} = {obj.value};'
            else:
                line = f'option {obj.name} = "{obj.value}";'

            if max < len(line):
                max = len(line)

        return max

    def option_string(self, obj: Option, max_length):
        if obj.value == 'true' or obj.value == 'false':
            line = f'option {obj.name} = {obj.value};'
        else:
            line = f'option {obj.name} = "{obj.value}";'

        return self.make_string(line, obj.comments, max_length)

    def imports_string(self, obj_list):
        if not obj_list:
            return ''

        max_length = self.max_length_of_import(obj_list)

        string_list = []
        for obj in obj_list:
            string = self.import_string(obj, max_length)
            string_list.append(string)
        return '\n'.join(string_list)

    def max_length_of_import(self, obj_list):
        max = 0
        for obj in obj_list:
            line = f'import "{obj.value}";'

            if max < len(line):
                max = len(line)

        return max

    def import_string(self, obj: Import, max_length):
        line = f'import "{obj.value}";'
        return self.make_string(line, obj.comments, max_length)

    def objects_string(self, obj_list):
        string_list = []
        for obj in obj_list:
            if isinstance(obj, Message):
                string = self.message_string(obj)
            if isinstance(obj, ProtoEnum):
                string = self.enum_string(obj)
            if isinstance(obj, Service):
                string = self.service_string(obj)
            string_list.append(string)
        return '\n\n'.join(string_list)

    def messages_string(self, obj_list):
        string_list = []
        for obj in obj_list:
            string = self.message_string(obj)
            string_list.append(string)
        return '\n\n'.join(string_list)

    def message_string(self, obj: Message):
        line = f'message {obj.name} ' + "{"
        message_header = self.make_string(line, obj.comments)
        elements_string = self.message_elemnents_string(obj.elements)
        s = '\n'.join([message_header, elements_string, '}'])
        return s

    def message_elemnents_string(self, obj_list):
        max_length = self.max_length_of_message_element(obj_list)

        string_list = []
        for obj in obj_list:
            if self.equal_sign:
                string = self.message_elemnent_string_align_with_equal_sign(obj, max_length)
            else:
                string = self.message_elemnent_string(obj, max_length)

            string_list.append(string)
        return '\n'.join(string_list)

    def message_elemnent_string(self, obj: MessageElement, max_length):
        if obj.label:
            line = f'{obj.label} {obj.type} {obj.name} = {obj.number};'
        else:
            line = f'{obj.type} {obj.name} = {obj.number};'

        return self.make_string(line, obj.comments, max_length, indents=self.indents)

    def message_elemnent_string_align_with_equal_sign(self, obj: MessageElement, max_length):
        if obj.label:
            line = f'{obj.label} {obj.type} {obj.name}'
        else:
            line = f'{obj.type} {obj.name}'

        need_fill_space_amount = max_length - len(line)
        line = f'{line}{" " * need_fill_space_amount} = {obj.number};'

        return self.make_string(line, obj.comments, max_length, indents=self.indents)

    def max_length_of_message_element(self, obj_list):
        max = 0
        for obj in obj_list:
            if self.equal_sign:
                if obj.label:
                    line = f'{obj.label} {obj.type} {obj.name}'
                else:
                    line = f'{obj.type} {obj.name}'
            else:
                if obj.label:
                    line = f'{obj.label} {obj.type} {obj.name} = {obj.number};'
                else:
                    line = f'{obj.type} {obj.name} = {obj.number};'

            if max < len(line):
                max = len(line)

        return max

    def enum_string(self, obj: ProtoEnum):
        line = f'enum {obj.name} ' + "{"
        message_header = self.make_string(line, obj.comments)
        elements_string = self.enum_elemnents_string(obj.elements)
        s = '\n'.join([message_header, elements_string, '}'])
        return s

    def enum_elemnents_string(self, obj_list):
        if self.equal_sign:
            max_length = self.max_length_of_enum_element_name(obj_list)
        else:
            max_length = self.max_length_of_enum_element(obj_list)

        string_list = []
        for obj in obj_list:
            string = self.enum_elemnent_string(obj, max_length)
            string_list.append(string)
        return '\n'.join(string_list)

    def enum_elemnent_string(self, obj: EnumElement, max_length):
        if self.equal_sign:
            line = f'{obj.name}'
            need_fill_space_amount = max_length - len(line)
            line = f'{line}{" " * need_fill_space_amount} = {obj.number};'
        else:
            line = f'{obj.name} = {obj.number};'

        return self.make_string(line, obj.comments, max_length, indents=self.indents)

    def max_length_of_enum_element(self, obj_list):
        max = 0
        for obj in obj_list:
            line = f'{obj.name} = {obj.number};'

            if max < len(line):
                max = len(line)

        return max

    def max_length_of_enum_element_name(self, obj_list):
        max = 0
        for obj in obj_list:
            line = f'{obj.name}'

            if max < len(line):
                max = len(line)

        return max

    def service_string(self, obj: ProtoEnum):
        line = f'service {obj.name} ' + "{"
        message_header = self.make_string(line, obj.comments)
        elements_string = self.service_elemnents_string(obj.elements)
        s = '\n'.join([message_header, elements_string, '}'])
        return s

    def service_elemnents_string(self, obj_list):
        max_length = self.max_length_of_service_element(obj_list)

        string_list = []
        for obj in obj_list:
            string = self.service_elemnent_string(obj, max_length)
            string_list.append(string)
        return '\n'.join(string_list)

    def service_elemnent_string(self, obj: ServiceElement, max_length):
        # rpc SeatAvailability (SeatAvailabilityRequest) returns (SeatAvailabilityResponse);
        line = f'{obj.label} {obj.name} ({obj.request}) returns ({obj.response});'

        return self.make_string(line, obj.comments, max_length, indents=self.indents)

    def max_length_of_service_element(self, obj_list):
        max = 0
        for obj in obj_list:
            line = f'{obj.label} {obj.name} ({obj.request}) returns ({obj.response});'

            if max < len(line):
                max = len(line)

        return max

    def make_string(self, value_line, comments, max_length=None, indents=0):
        lines = []
        top_comment_lines = []
        right_comment = ''
        if self.all_top_comments:
            for comment in comments:
                if comment.position == Position.TOP:
                    text_lines = [l.strip() for l in comment.text.split('\n')]
                    top_comment_lines.extend(text_lines)
                if comment.position == Position.Right:
                    top_comment_lines.append(comment.text)
                    right_comment = ''
        else:
            for comment in comments:
                if comment.position == Position.TOP:
                    text_lines = [l.strip() for l in comment.text.split('\n')]
                    top_comment_lines.extend(text_lines)
                if comment.position == Position.Right:
                    right_comment = comment.text

        if top_comment_lines:
            top_comment_lines.insert(0, '/*')
            top_comment_lines.append('*/')

        lines = top_comment_lines
        if right_comment:
            if max_length is not None:
                need_fill_space_amount = max_length - len(value_line)
                line = f'{value_line}{" " * need_fill_space_amount}  // {right_comment}'
            else:
                line = f'{value_line}  // {right_comment}'
            lines.append(line)
        else:
            lines.append(value_line)

        indented_lines = [f'{" " * indents}{line}' for line in lines]
        string = '\n'.join(indented_lines)
        return string



from enum import Enum
from proto_formatter.constant import Constant
from proto_formatter.comment import CommentParser
from proto_formatter.object_parser import ObjectParser
from proto_formatter.proto import ProtoEnum, EnumElement
from proto_formatter.util import remove_prefix, remove_suffix


class EnumParser(ObjectParser):
    def __init__(self):
        obj = ProtoEnum(name='name', elements=[], comments=[])
        super(EnumParser, self).__init__(obj=obj)

    @classmethod
    def parse_obj_field(cls, line, top_comments=[]):
        line = line.strip()
        equal_sign_index = line.index(cls.EQUAL_SIGN)
        semicolon_index = line.index(cls.SEMICOLON)
        str_before_equqal_sign = line[:equal_sign_index]
        parts = str_before_equqal_sign.split(' ')
        parts = list(filter(None, parts))
        value = line[equal_sign_index + 1:semicolon_index].strip().replace('"', "").replace("'", "")
        comments = CommentParser.create_comment(line, top_comments)
        return EnumElement(name=parts[0], number=value, comments=comments)
